#!/usr/bin/python

# An X11 desktop panel that contains launchers for items in a given XDG Menu
# Copyright (C) 2010 Michael Hall <mhall119@ubuntu.com>
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# python modules
import os
import sys
import gtk
import glib
import gobject
import gmenu
import subprocess
import math
from threading import Lock

class XDGLauncher:
    
    def __init__(self, menu_name, menu_file=None):
        self.__tree = menu_file or 'applications.menu'
        self.__rebuild_lock = Lock()
        self.__schedule_id = {self.__tree: None,}
        self.__schedule_lock = Lock()
        self.__menu_rebuild_delay = 2
        self.__menu_name = menu_name

        # Variables & config
        self.bar_width = self.bar_height =  0
        self.bar_resize_x = self.bar_resize_y = 0
        self.bar_hide_x = self.bar_hide_y = 0

        self.bar_hidden = False
        # timer for leave bar callback
        self.timer_auto_hide = None
        # timer for animate hidding
        self.timer_smooth_hide = None

        self.timer = 2
        self.auto_hide = True
        self.offset_pos = 5
        self.auto_raise = True

        self.smooth_hide = True

        # use for animate hiding
        self.timer_anim = None
        self.moving = False
        self.count = 0
        self.countdown = 0

        self.hidden_size = 2
        self.position = "right"

        self.padding = 5
        self.icon_size = 90

        launchers = []

        self.win = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.win.set_title("XDG Launcher")
        self.win.set_decorated(False)
        self.win.set_skip_taskbar_hint(True)
        self.win.set_skip_pager_hint(True)
        self.win.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_DOCK)
        self.win.stick()
        self.win.set_app_paintable(True)
        self.win.add_events(gtk.gdk.BUTTON_PRESS_MASK|gtk.gdk.ENTER_NOTIFY|gtk.gdk.LEAVE_NOTIFY)
        self.win.connect("button-press-event", self.bar_released)
        self.win.connect("leave-notify-event", self.bar_leave_notify)
        self.win.connect("enter-notify-event", self.bar_enter_notify)
        #self.win.connect('screen-changed', self.screen_changed)
        self.win.set_border_width(0)

        self.container = gtk.HBox()
        self.container.set_homogeneous(False)
        self.container.set_border_width(4)
        self.container.set_spacing(4)

        self.theme = gtk.icon_theme_get_default()
        self.running = dict()
        self.build_menu(self.__tree)
        self.win.add(self.container)
        self.create_menu()
        gobject.timeout_add(10, self.reposition)
        self.toggle_hidden()

    def create_menu(self):
        self.popupMenu = gtk.Menu()
        menuPopup = gtk.MenuItem ("Position")
        self.menu_position = gtk.Menu()
        self.pos_top = gtk.RadioMenuItem(None, 'Top')
        self.pos_bottom = gtk.RadioMenuItem(None, 'Bottom')
        self.pos_left = gtk.RadioMenuItem(None, 'Left')
        self.pos_right = gtk.RadioMenuItem(None, 'Right')
        #self.by_group.set_active(config.b_order_by_group)
        #self.by_status.set_active(not config.b_order_by_group)
        self.menu_position.append(self.pos_top)
        self.menu_position.append(self.pos_bottom)
        self.menu_position.append(self.pos_left)
        self.menu_position.append(self.pos_right)
        menuPopup.set_submenu (self.menu_position)
        self.popupMenu.add(menuPopup)
        menuPopup = gtk.ImageMenuItem (gtk.STOCK_ABOUT)
        #menuPopup.connect("activate", self.show_about_dialog)
        self.popupMenu.add(menuPopup)
        menuPopup = gtk.ImageMenuItem (gtk.STOCK_QUIT)
        #menuPopup.connect("activate", self.doquit)
        self.popupMenu.add(menuPopup)
        self.popupMenu.show_all()

    def on_timeout_hide(self):
        self.countdown -= 1
        N = self.count
        n = self.countdown
        x = self.bar_resize_x + ( (self.bar_hide_x - self.bar_resize_x) / N ) * (N-n)

        if self.position == "top":
            y =  ( (self.bar_hide_y - self.bar_resize_y ) / N ) * (N-n)
        else:
            y = self.bar_resize_y + ( (self.bar_hide_y - self.bar_resize_y ) / N ) * (N-n)

        self.win.move(x, y)
        if self.countdown <= 0:
            self.bar_hidden = True
            self.bar_resize()
            self.moving = False
            return False
        else:
            return True

    def bar_leave_notify(self, widget=None, event=None):
        if not self.timer_auto_hide == None:
            gobject.source_remove(self.timer_auto_hide)
            self.timer_auto_hide = None

        if self.smooth_hide:
            if self.timer == 0:
                if self.offset_pos > 0:
                    ## minimum time because bar auto-hide it-self :(
                    self.timer_auto_hide = gobject.timeout_add(1000, self.on_timeout_notify)
                else:
                    ## minimum time because bar auto-hide it-self :(
                    self.timer_auto_hide = gobject.timeout_add(250, self.on_timeout_notify)
            else:
                self.timer_auto_hide = gobject.timeout_add(self.timer*1000, self.on_timeout_notify)
            self.focus = None
            self.update()

        if not self.timer_anim == None:
            gobject.source_remove(self.timer_anim)

    def bar_enter_notify(self, widget=None, event=None):
        if self.auto_raise and self.bar_hidden:
            self.toggle_hidden()

        if not self.timer_auto_hide == None:
            gobject.source_remove(self.timer_auto_hide)
            self.timer_auto_hide = None

        if self.moving:
            self.moving = False
            self.bar_hidden = True
            gobject.source_remove(self.timer_smooth_hide)
            self.timer_smooth_hide = None
            self.toggle_hidden()

    def on_timeout_notify(self):
        if not self.bar_hidden:
            self.toggle_hidden()
            gobject.source_remove(self.timer_auto_hide)
            self.timer_auto_hide = None
        return False

    def toggle_hidden(self, widget=None , event=None):
        if not self.auto_hide:
            return

        if  self.bar_hidden:
            self.bar_hidden = False
            self.bar_resize()
        else:
            self.bar_hide()

    def update(self):
        self.win.queue_draw()
        return True

    def bar_resize(self):
        if self.bar_hidden:
            self.win.move(self.bar_hide_x, self.bar_hide_y)
        else:
            self.win.move(self.bar_resize_x, self.bar_resize_y)

    def position_panel(self):

        self.bar_width , self.bar_height = self.win.get_size()
        screen_width = gtk.gdk.screen_width()
        screen_height = gtk.gdk.screen_height()

        if self.position == "bottom":
            self.disposition = "Horizontal"
            self.bar_resize_y = screen_height - self.bar_height - self.offset_pos
            self.bar_resize_x = ( screen_width - self.bar_width ) / 2
            self.bar_hide_y = screen_height - self.hidden_size
            self.bar_hide_x = self.bar_resize_x

        if self.position == "top":
            self.disposition = "Horizontal"
            self.bar_resize_y = 0 + self.offset_pos
            self.bar_resize_x = ( screen_width - self.bar_width ) / 2
            self.bar_hide_y = self.hidden_size - self.bar_height
            self.bar_hide_x = self.bar_resize_x

        if self.position == "left":
            self.disposition = "Vertical"
            self.bar_resize_x = self.offset_pos - 4
            self.bar_resize_y = (screen_height - self.bar_height) / 2
            self.bar_hide_y = self.bar_resize_y
            self.bar_hide_x = - self.bar_width + self.hidden_size

        if self.position == "right":
            self.disposition = "Vertical"
            self.bar_resize_x = screen_width - self.bar_width
            self.bar_resize_y = (screen_height - self.bar_height) / 2
            self.bar_hide_y = self.bar_resize_y
            self.bar_hide_x = screen_width - self.hidden_size


        print "Screen (WxH): %sx%s" % (screen_width,screen_height)
        print "Launcher (WxH): %sx%s" % (self.bar_width,self.bar_height)
        print "Disposition : %s" % self.disposition

        

    def bar_released(self, widget, event):
        if event.button==3: # right click
            self.popupMenu.popup(None, None, None, event.button, event.time)

        elif event.button==2: # middle click
            self.auto_hide = not self.auto_hide
            if self.bar_hidden:
                self.toggle_hidden()
                pass

        elif event.button==1 and self.bar_hidden: # left click
            self.toggle_hidden()

    def bar_hide(self):
        if self.smooth_hide:
            T = 30 # ms
            self.count = 500 / T
            self.countdown = self.count
            self.moving = True
            self.timer_smooth_hide = gobject.timeout_add(T, self.on_timeout_hide)
        else:
            self.bar_hidden = True
            self.bar_resize()

    def reposition(self):
        self.position_panel()
        self.bar_resize()
        self.update()
        return False

    def build_menu(self, tree, rebuild=False):
        import math
        launchers = []
        if not rebuild:
            menu = gmenu.lookup_tree(self.__tree)
            menu.add_monitor(self.menu_changed_cb, launchers)
            if not self.__menu_name.startswith('/'):
               self.__menu_name = '/%s'%self.__menu_name
            entries = menu.get_directory_from_path(self.__menu_name)
        else:
            entries = tree.get_directory_from_path(self.__menu_name)

        for e in entries.get_contents():
            if isinstance(e, gmenu.Entry):
                launchers.append(e)

        for child in self.container.get_children():
            self.container.remove(child)

        col = 0
        row = 0
        self.b_box=gtk.Toolbar()
        self.b_box.connect("button-press-event", self.bar_released)
        if self.position == 'bottom' or self.position == 'top':
            self.b_box.set_orientation(gtk.ORIENTATION_HORIZONTAL)
        elif self.position == 'left' or self.position == 'right':
            self.b_box.set_orientation(gtk.ORIENTATION_VERTICAL)
        self.b_box.set_style(gtk.TOOLBAR_ICONS)
        for app in launchers:
            icon = gtk.Image()
            if os.path.isfile(app.icon):
                icon.set_from_file(app.icon)
            if os.path.isfile(os.path.join('/usr/share/pixmaps', app.icon)):
                icon.set_from_file(os.path.join('/usr/share/pixmaps', app.icon))
            else:
                try:
                    img = self.theme.load_icon(app.icon, self.icon_size, gtk.ICON_LOOKUP_FORCE_SIZE)
                except:
                    img = self.theme.load_icon('application-default-icon', self.icon_size, gtk.ICON_LOOKUP_FORCE_SIZE)
                icon.set_from_pixbuf(img)

            b = gtk.Button()
            b.props.relief = gtk.RELIEF_NONE
            b.set_image(icon)
            b.set_border_width(0)
            b.set_can_focus(False)
            b.set_name(app.name)
            b.set_tooltip_text('%s\n%s' % (app.name, app.comment))
            b.connect('clicked', self.launch_app_handler, app.exec_info)
            b.set_size_request(self.icon_size+15, self.icon_size+15)
            b.show()
            b.connect("button-press-event", self.bar_released)
            
            self.b_box.add(b)
            self.b_box.show()
            self.container.pack_start(self.b_box,expand=False)
            b = None
        if rebuild:
            self.reposition()


    def launch_app_handler(self, widget, exe):
        if not self.running.has_key(exe) or self.running[exe].poll() is not None:
            self.running[exe] = subprocess.Popen([exe])

    def get_bar_width_ratio(self, numIcons):
        if numIcons < 2:
            return 0.1
        if numIcons < 3:
            return 0.2
        return 1 - (math.sqrt(numIcons) / numIcons)

    def menu_changed_cb(self, menu_tree, menu_items):
        def refresh_menu(tree, items):
            with self.__rebuild_lock:
                # Redraw the menu
                self.build_menu(tree, True)
            return False
        with self.__schedule_lock:
            file = menu_tree.menu_file
            if self.__schedule_id[file] is not None:
                glib.source_remove(self.__schedule_id[file])
            self.__schedule_id[file] = glib.timeout_add_seconds(self.__menu_rebuild_delay, refresh_menu, menu_tree, menu_items)

if __name__ == "__main__":
    if sys.argv[0] == __file__:
        if len(sys.argv) > 1:
            menu = sys.argv[1]
        else:
            menu = None
    else:
        if len(sys.argv) > 0:
            menu = sys.argv[0]
        else:
            menu = None
        
    l = XDGLauncher(menu or 'Games')
    l.win.show_all()
    #l.win.window.property_change("_NET_WM_STRUT", "CARDINAL", 32, gtk.gdk.PROP_MODE_REPLACE, [0,0,0,l.bar_height])

    gtk.gdk.threads_enter()
    gtk.main()
    gtk.gdk.threads_leave()
